""" This module handles TCP connections to the OpenWebNet gateway """

import asyncio
import hmac
import hashlib
import string 
import random
import logging
import traceback
from urllib.parse import urlparse

from .discovery import find_gateways, get_gateway, get_port
from .message import OWNMessage, OWNSignaling, OWNEvent


class OWNGateway():

    def __init__(self, discovery_info: dict):
        # Attributes potentially provided by user
        self.address = discovery_info["address"] if "address" in discovery_info else None
        self._password = discovery_info["password"] if "password" in discovery_info else None
        # Attributes retrieved from SSDP discovery
        self.ssdp_location = discovery_info["ssdp_location"] if "ssdp_location" in discovery_info else None
        self.ssdp_st = discovery_info["ssdp_st"] if "ssdp_st" in discovery_info else None
        # Attributes retrieved from UPnP device description
        self.deviceType = discovery_info["deviceType"] if "deviceType" in discovery_info else None
        self.friendlyName = discovery_info["friendlyName"] if "friendlyName" in discovery_info else None
        self.manufacturer = discovery_info["manufacturer"] if "manufacturer" in discovery_info else None
        self.manufacturerURL = discovery_info["manufacturerURL"] if "manufacturerURL" in discovery_info else None
        self.modelName = discovery_info["modelName"] if "modelName" in discovery_info else None
        self.modelNumber = discovery_info["modelNumber"] if "modelNumber" in discovery_info else None
        #self.presentationURL = discovery_info["presentationURL"] if "presentationURL" in discovery_info else None
        self.serialNumber = discovery_info["serialNumber"] if "serialNumber" in discovery_info else None
        self.UDN = discovery_info["UDN"] if "UDN" in discovery_info else None
        # Attributes retrieved from SOAP service control
        self.port = discovery_info["port"] if "port" in discovery_info else None

    @property
    def id(self) -> str:
        return self.serialNumber
    
    @id.setter
    def id(self, id: str) -> None:
        self.serialNumber = id

    @property
    def host(self) -> str:
        return self.address

    @host.setter
    def host(self, host: str) -> None:
        self.address = host

    @property
    def firmware(self) -> str:
        return self.modelNumber
    
    @firmware.setter
    def firmware(self, firmware: str) -> None:
        self.modelNumber = firmware

    @property
    def serial(self) -> str:
        return self.serialNumber

    @serial.setter
    def serial(self, serial: str) -> None:
        self.serialNumber = serial

    @property
    def password(self) -> str:
        return self._password

    @password.setter
    def password(self, password: str) -> None:
        self._password = password

    @classmethod
    async def get_first_available_gateway(cls, password: str = None):
        local_gateways = await find_gateways()
        local_gateways[0]["password"] = password
        return cls(local_gateways[0])
    
    @classmethod
    async def find_from_address(cls, address: str):
        if address is not None:
            return cls(await get_gateway(address))
        else:
            return await cls.get_first_available_gateway()

    @classmethod
    async def build_from_discovery_info(cls, discovery_info: dict):

        if (("address" not in discovery_info or discovery_info["address"] is None) and "ssdp_location" in discovery_info and discovery_info["ssdp_location"] is not None):
            discovery_info["address"] = urlparse(discovery_info["ssdp_location"]).hostname
        
        if "port" in discovery_info and discovery_info["port"] is None:
            if "ssdp_location" in discovery_info and discovery_info["ssdp_location"] is not None:
                discovery_info["port"] = await get_port(discovery_info["ssdp_location"])
            elif "address" in discovery_info and discovery_info["address"] is not None:
                return await cls.find_from_address(discovery_info["address"])
            else:
                return await cls.get_first_available_gateway(password = discovery_info["password"] if "password" in discovery_info else None)

        return cls(discovery_info)

class OWNSession():
    """ Connection to OpenWebNet gateway """

    SEPARATOR = '##'.encode()

    def __init__(self, gateway: OWNGateway = None, connection_type: str = "test", logger: logging.Logger = None):
        """ Initialize the class
        Arguments:
        logger: instance of logging
        address: IP address of the OpenWebNet gateway
        port: TCP port for the connection
        password: OpenWebNet password
        """
        
        self._gateway = gateway
        self._type = connection_type.lower()
        self._logger = logger

        self._stream_reader : asyncio.StreamReader
        self._stream_writer : asyncio.StreamWriter

    @property
    def gateway(self) -> OWNGateway:
        return self._gateway

    @gateway.setter
    def gateway(self, gateway: OWNGateway) -> None:
        self._gateway = gateway

    @property
    def password(self) -> str:
        return str(self._password)

    @password.setter
    def password(self, password: str) -> None:
        self._password = password

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @logger.setter
    def logger(self, logger: logging.Logger) -> None:
        self._logger = logger

    @property
    def connection_type(self) -> str:
        return self._type

    @connection_type.setter
    def connection_type(self, connection_type: str) -> None:
        self._type = connection_type.lower()

    @classmethod
    async def test_gateway(cls, gateway: OWNGateway) -> dict:
        connection = cls(gateway)
        return await connection.test_connection()

    async def test_connection(self) -> dict:
    
        retry_count = 0
        retry_timer = 1

        while True:
            try:
                if retry_count > 2:
                    self._logger.error("Test session connection still refused after 3 attempts.")
                    return None
                self._stream_reader, self._stream_writer = await asyncio.open_connection(self._gateway.address, self._gateway.port)
                break
            except ConnectionRefusedError:
                self._logger.warning("Test session connection refused, retrying in %ss.", retry_timer)
                await asyncio.sleep(retry_timer)
                retry_count += 1
                retry_timer *= 2
        
        try:
            result = await self._negotiate()
            await self.close()
        except ConnectionResetError:
            error = True
            error_message = "password_retry"
            self._logger.error("Negotiation reset while opening %s session. Wait 60 seconds before retrying.", self._type)

            return {"Success": not error, "Message": error_message}
            
        return result

    async def connect(self):
        self._logger.debug("Opening %s session.", self._type)
        
        retry_count = 0
        retry_timer = 1

        while True:
            try:
                if retry_count > 4:
                    self._logger.error("%s session connection still refused after 5 attempts.", self._type.capitalize())
                    return None
                self._stream_reader, self._stream_writer = await asyncio.open_connection(self._gateway.address, self._gateway.port)
                return await self._negotiate()
            except (ConnectionRefusedError, asyncio.IncompleteReadError):
                self._logger.warning("%s session connection refused, retrying in %ss.", self._type.capitalize(), retry_timer)
                await asyncio.sleep(retry_timer)
                retry_count += 1
                retry_timer = retry_count*2
            except ConnectionResetError:
                self._logger.warning("%s session connection reset, retrying in 60s.", self._type.capitalize())
                await asyncio.sleep(60)
                retry_count += 1

    async def close(self) -> None:
        """ Closes the connection to the OpenWebNet gateway """
        self._stream_writer.close()
        await self._stream_writer.wait_closed()
        self._logger.debug("%s session closed.", self._type.capitalize())
    
    async def _negotiate(self) -> dict:

        type_id = 0 if self._type == "command" else 1
        error = False
        error_message = None

        self._logger.debug("Negotiating %s session.", self._type)

        self._stream_writer.write(f"*99*{type_id}##".encode())
        await self._stream_writer.drain()

        raw_response = await self._stream_reader.readuntil(OWNSession.SEPARATOR)
        resulting_message = OWNSignaling(raw_response.decode())
        self._logger.debug("Reply: %s", resulting_message)

        if resulting_message.is_NACK():
            self._logger.error("Error while opening %s session.", self._type)
            error = True
            error_message = "connection_refused"

        raw_response = await self._stream_reader.readuntil(OWNSession.SEPARATOR)
        resulting_message = OWNSignaling(raw_response.decode())
        if resulting_message.is_NACK():
            error = True
            error_message = "negotiation_refused"
            self._logger.debug("Reply: %s", resulting_message)
            self._logger.error("Error while opening %s session.", self._type)
        elif resulting_message.is_SHA():
            self._logger.debug("Received SHA challenge: %s", resulting_message)
            if self._gateway.password is None:
                    error = True
                    error_message = "password_required"
                    self._logger.warning("Connection requires a password but none was provided.")
                    self._stream_writer.write("*#*0##".encode())
                    await self._stream_writer.drain()
            else:
                if resulting_message.is_SHA_1():
                    self._logger.debug("Detected SHA-1 method.")
                    method = "sha1"
                elif resulting_message.is_SHA_256():
                    self._logger.debug("Detected SHA-256 method.")
                    method = "sha256"
                self._logger.debug("Accepting challenge, initiating handshake.")
                self._stream_writer.write("*#*1##".encode())
                await self._stream_writer.drain()
                raw_response = await self._stream_reader.readuntil(OWNSession.SEPARATOR)
                resulting_message = OWNSignaling(raw_response.decode())
                if resulting_message.is_nonce():
                    ra = resulting_message.nonce
                    self._logger.debug("Received Ra.")
                    key = ''.join(random.choices(string.digits, k = 56))
                    rb = self._hex_string_to_int_string(hmac.new(key=key.encode(), digestmod=method).hexdigest())
                    self._logger.debug("Generated Rb.")
                    hashedPass = f"*#{rb}*{self._encode_hmac_password(method=method, password=self._gateway.password, nonce_a=ra, nonce_b=rb)}##"
                    self._logger.debug("Sending %s session password.", self._type)
                    self._stream_writer.write(hashedPass.encode())
                    await self._stream_writer.drain() 
                    try:
                        raw_response = await asyncio.wait_for(self._stream_reader.readuntil(OWNSession.SEPARATOR), timeout=5)
                        resulting_message = OWNSignaling(raw_response.decode())
                        if resulting_message.is_NACK():
                            error = True
                            error_message = "password_error"
                            self._logger.error("Password error while opening %s session.", self._type)
                        elif resulting_message.is_nonce():
                            self._logger.debug("Received HMAC response.")
                            hmac_response = resulting_message.nonce
                            if hmac_response == self._decode_hmac_response(method=method, password=self._gateway.password, nonce_a=ra, nonce_b=rb):
                                self._logger.debug("Server identity confirmed.")
                                self._stream_writer.write("*#*1##".encode())
                                await self._stream_writer.drain()
                            else:
                                self._logger.error("Server identity could not be confirmed.")
                                self._stream_writer.write("*#*0##".encode())
                                await self._stream_writer.drain()
                                error = True
                                error_message = "negociation_error"
                                self._logger.error("Error while opening %s session: HMAC authentication failed.", self._type)
                    except asyncio.IncompleteReadError:
                        error = True
                        error_message = "password_error"
                        self._logger.error("Password error while opening %s session.", self._type)
                    except asyncio.TimeoutError:
                        error = True
                        error_message = "password_error"
                        self._logger.error("Password timeout error while opening %s session.", self._type)
        elif resulting_message.is_nonce():
            self._logger.debug("Received nonce: %s", resulting_message)
            if self._gateway.password is not None:
                hashedPass = f"*#{self._get_own_password(self._gateway.password, resulting_message.nonce)}##"
                self._logger.debug("Sending %s session password.", self._type)
                self._stream_writer.write(hashedPass.encode())
                await self._stream_writer.drain()
                raw_response = await self._stream_reader.readuntil(OWNSession.SEPARATOR)
                resulting_message = OWNSignaling(raw_response.decode())
                self._logger.debug("Reply: %s", resulting_message)
                if resulting_message.is_NACK():
                    error = True
                    error_message = "password_error"
                    self._logger.error("Password error while opening %s session.", self._type)
                elif resulting_message.is_ACK():
                    self._logger.debug("%s session established.", self._type.capitalize())
            else:
                error = True
                error_message = "password_error"
                self._logger.error("Connection requires a password but none was provided while opening %s session.", self._type)
        elif resulting_message.is_ACK():
            self._logger.debug("Reply: %s", resulting_message)
            self._logger.debug("%s session established.", self._type.capitalize())
        else:
            error = True
            error_message = "negotiation_failed"
            self._logger.debug("Unexpected message during negotiation: %s", resulting_message)

        return {"Success": not error, "Message": error_message}

    def _get_own_password(self, password, nonce, test=False):
        start = True    
        num1 = 0
        num2 = 0
        password = int(password)
        if test:
            print("password: %08x" % (password))
        for c in nonce :
            if c != "0":
                if start:
                    num2 = password
                start = False
            if test:
                print("c: %s num1: %08x num2: %08x" % (c, num1, num2))
            if c == '1':
                num1 = (num2 & 0xFFFFFF80) >> 7
                num2 = num2 << 25
            elif c == '2':
                num1 = (num2 & 0xFFFFFFF0) >> 4
                num2 = num2 << 28
            elif c == '3':
                num1 = (num2 & 0xFFFFFFF8) >> 3
                num2 = num2 << 29
            elif c == '4':
                num1 = num2 << 1
                num2 = num2 >> 31
            elif c == '5':
                num1 = num2 << 5
                num2 = num2 >> 27
            elif c == '6':
                num1 = num2 << 12
                num2 = num2 >> 20
            elif c == '7':
                num1 = num2 & 0x0000FF00 | (( num2 & 0x000000FF ) << 24 ) | (( num2 & 0x00FF0000 ) >> 16 )
                num2 = ( num2 & 0xFF000000 ) >> 8
            elif c == '8':
                num1 = (num2 & 0x0000FFFF) << 16 | ( num2 >> 24 )
                num2 = (num2 & 0x00FF0000) >> 8
            elif c == '9':
                num1 = ~num2
            else :
                num1 = num2
    
            num1 &= 0xFFFFFFFF
            num2 &= 0xFFFFFFFF
            if (c not in "09"):
                num1 |= num2
            if test:
                print("     num1: %08x num2: %08x" % (num1, num2))
            num2 = num1
        return num1

    def _encode_hmac_password(self, method: str, password: str, nonce_a: str, nonce_b: str):
        if method == 'sha1':
            message = self._int_string_to_hex_string(nonce_a) + self._int_string_to_hex_string(nonce_b) + "736F70653E" + "636F70653E" + hashlib.sha1(password.encode()).hexdigest()
            return self._hex_string_to_int_string(hashlib.sha1(message.encode()).hexdigest())
        elif method == 'sha256':
            message = self._int_string_to_hex_string(nonce_a) + self._int_string_to_hex_string(nonce_b) + "736F70653E" + "636F70653E" + hashlib.sha256(password.encode()).hexdigest()
            return self._hex_string_to_int_string(hashlib.sha256(message.encode()).hexdigest())
        else:
            return None

    def _decode_hmac_response(self, method: str, password: str, nonce_a: str, nonce_b: str):
        if method == 'sha1':
            message = self._int_string_to_hex_string(nonce_a) + self._int_string_to_hex_string(nonce_b) + hashlib.sha1(password.encode()).hexdigest()
            return self._hex_string_to_int_string(hashlib.sha1(message.encode()).hexdigest())
        elif method == 'sha256':
            message = self._int_string_to_hex_string(nonce_a) + self._int_string_to_hex_string(nonce_b) + hashlib.sha256(password.encode()).hexdigest()
            return self._hex_string_to_int_string(hashlib.sha256(message.encode()).hexdigest())
        else:
            return None

    def _int_string_to_hex_string(self, int_string: str) -> str:
        hex_string = ""
        for i in range(0, len(int_string), 2):
            hex_string += f"{int(int_string[i:i+2]):x}"
        return hex_string

    def _hex_string_to_int_string(self, hex_string: str) -> str:
        int_string = ""
        for i in range(0, len(hex_string), 1):
            int_string += f"{int(hex_string[i:i+1], 16):0>2d}"
        return int_string

class OWNEventSession(OWNSession):

    def __init__(self, gateway: OWNGateway = None, logger: logging.Logger = None):
        super().__init__(gateway = gateway, connection_type = "event", logger = logger)

    @classmethod
    async def connect_to_gateway(cls, gateway: OWNGateway):
        connection = cls(gateway)
        await connection.connect()
    
    async def get_next(self):
        """ Acts as an entry point to read messages on the event bus.
        It will read one frame and return it as an OWNMessage object """
        try:
            data = await self._stream_reader.readuntil(OWNSession.SEPARATOR)
            return OWNMessage.parse(data.decode())
        except asyncio.IncompleteReadError as ire:
            self._logger.warning("Connection interrupted, reconnecting...")
            #traceback.print_tb(ire.__traceback__)
            await self.connect()
            return None
        except AttributeError as ar:
            self._logger.exception("Received data could not be parsed into a message:")
            #traceback.print_tb(ar.__traceback__)
            return None
        except ConnectionError as e:
            self._logger.exception("Connection error:")
            #traceback.print_tb(e.__traceback__)
            return None
        except Exception as e:
            self._logger.exception("Event session crashed.")
            #traceback.print_tb(e.__traceback__)
            return None

class OWNCommandSession(OWNSession):

    def __init__(self, gateway: OWNGateway = None, logger: logging.Logger = None):
        super().__init__(gateway = gateway, connection_type = "command", logger = logger)

    @classmethod
    async def send_to_gateway (cls,  message: str, gateway: OWNGateway):
        connection = cls(gateway)
        await connection.connect()
        await connection.send(message)

    @classmethod
    async def connect_to_gateway(cls, gateway: OWNGateway):
        connection = cls(gateway)
        await connection.connect()

    async def send(self, message: str, is_status_request: bool = False):
        """ Send the attached message on an existing 'command' connection,
        actively reconnecting it if it had been reset. """

        try:
            self._stream_writer.write(str(message).encode())
            await self._stream_writer.drain()
            raw_response = await self._stream_reader.readuntil(OWNSession.SEPARATOR)
            resulting_message = OWNMessage.parse(raw_response.decode())
            if isinstance(resulting_message, OWNSignaling) and resulting_message.is_NACK():
                self._stream_writer.write(str(message).encode())
                await self._stream_writer.drain()
                raw_response = await self._stream_reader.readuntil(OWNSession.SEPARATOR)
                resulting_message = OWNSignaling(raw_response.decode())
                if resulting_message.is_NACK():
                    self._logger.error("Could not send message %s.", message)
                elif resulting_message.is_ACK():
                    if not is_status_request:
                        self._logger.info("Message %s was successfully sent.", message)
                    else:
                        self._logger.debug("Message %s was successfully sent.", message)
            elif isinstance(resulting_message, OWNSignaling) and resulting_message.is_ACK():
                if not is_status_request: 
                    self._logger.info("Message %s was successfully sent.", message)
                else:
                    self._logger.debug("Message %s was successfully sent.", message)
            else:
                self._logger.debug("Message %s received response %s.", message,  resulting_message)
                raw_response = await self._stream_reader.readuntil(OWNSession.SEPARATOR)
                resulting_message = OWNSignaling(raw_response.decode())
                if resulting_message.is_NACK():
                    self._logger.error("Could not send message %s.", message)
                elif resulting_message.is_ACK():
                    if not is_status_request:
                        self._logger.info("Message %s was successfully sent.", message)
                    else:
                        self._logger.debug("Message %s was successfully sent.", message)

        except (ConnectionResetError, asyncio.IncompleteReadError):
                self._logger.debug("Command session connection reset, retrying...")
                await self.connect()
                await self.send(message=message, is_status_request=is_status_request)
        except Exception as e:
            self._logger.exception("Command session crashed.")
            #traceback.print_tb(ex.__traceback__)
            return None


import argparse
import asyncio
import logging
import time

from .connection import (OWNEventSession, OWNGateway)


async def main(arguments: dict, connection: OWNEventSession) -> None:
    """ Package entry point! """
    
    address = arguments["address"] if "address" in arguments and isinstance(arguments["address"], str) else None
    port = arguments["port"] if "port" in arguments and isinstance(arguments["port"], int) else None
    password = arguments["password"] if "password" in arguments and isinstance(arguments["password"], str) else None
    serialNumber = arguments["serialNumber"] if "serialNumber" in arguments and isinstance(arguments["serialNumber"], str) else None
    logger = arguments["logger"] if "logger" in arguments and isinstance(arguments["logger"], logging.Logger) else None

    gateway = await OWNGateway.build_from_discovery_info({"address": address, "port": port, "password": password, "serialNumber": serialNumber})
    connection.gateway = gateway
    
    if logger is not None:
        connection.logger = logger

    await connection.connect()

    while True:
        message = await connection.get_next()
        if message:
            logger.debug("Received: %s", message)
            if message.is_event:
                logger.info(message.human_readable_log)

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--address", type=str, help="IP address of the OpenWebNet gateway")
    parser.add_argument("-p", "--port", type=int, help="TCP port to connectect the gateway, default is 20000")
    parser.add_argument("-P", "--password", type=str, help="Numeric password for the OpenWebNet connection, default is 12345")
    parser.add_argument("-m", "--mac", type=str, help="MAC address of the gateway (to be used as ID, if  not found via SSDP)")
    parser.add_argument("-v", "--verbose", type=int, help="Change output verbosity [0 = WARNING; 1 = INFO (default); 2 = DEBUG]")
    args = parser.parse_args()

    # create logger with 'OWNd'
    logger = logging.getLogger('OWNd')
    logger.setLevel(logging.DEBUG)

    # create console handler which logs even debug messages
    log_stream_handler = logging.StreamHandler()

    if args.verbose == 2:
        log_stream_handler.setLevel(logging.DEBUG)
    elif args.verbose == 0:
        log_stream_handler.setLevel(logging.WARNING)
    else:
        log_stream_handler.setLevel(logging.INFO)
    
    # create formatter and add it to the handlers
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    log_stream_handler.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(log_stream_handler)

    event_session = OWNEventSession()
    arguments = {"address": args.address, "port": args.port, "password": args.password, "serialNumber": args.mac, "logger": logger}

    loop = asyncio.get_event_loop()
    main_task = asyncio.ensure_future(main(arguments, event_session))
    #loop.set_debug(True)

    try:
        logger.info("Starting OWNd.")
        loop.run_forever()
        #asyncio.run(main(arguments))
    except KeyboardInterrupt:
        logger.info("Stoping OWNd.")
        main_task.cancel()
        loop.run_until_complete(event_session.close())
        loop.stop()
        loop.close()
    finally:
        logger.info('OWNd stopped.')


import asyncio
import email.parser
import errno
import logging
import socket
import xml.dom.minidom
from urllib.parse import urlparse

import aiohttp

class SSDPMessage:
    """Simplified HTTP message to serve as a SSDP message."""

    def __init__(self, version="HTTP/1.1", headers=None):
        if headers is None:
            headers = []
        elif isinstance(headers, dict):
            headers = headers.items()

        self.version = version
        self.headers = list(headers)
        self.headers_dictionary = {}
        for header in self.headers:
            self.headers_dictionary.setdefault(header[0], header[1])

    @classmethod
    def parse(cls, msg):
        """
        Parse message a string into a :class:`SSDPMessage` instance.
        Args:
            msg (str): Message string.
        Returns:
            SSDPMessage: Message parsed from string.
        """
        raise NotImplementedError()

    @classmethod
    def parse_headers(cls, msg):
        """
        Parse HTTP headers.
        Args:
            msg (str): HTTP message.
        Returns:
            (List[Tuple[str, str]]): List of header tuples.
        """
        return list(email.parser.Parser().parsestr(msg).items())

    def __str__(self):
        """Return full HTTP message."""
        raise NotImplementedError()

    def __bytes__(self):
        """Return full HTTP message as bytes."""
        _bytes = self.__str__().encode().replace(b"\n", b"\r\n")
        _bytes = _bytes + b"\r\n\r\n"
        return _bytes

class SSDPResponse(SSDPMessage):
    """Simple Service Discovery Protocol (SSDP) response."""

    def __init__(self, status_code, reason, **kwargs):
        self.status_code = int(status_code)
        self.reason = reason
        super().__init__(**kwargs)

    @classmethod
    def parse(cls, msg):
        """Parse message string to response object."""
        lines = msg.splitlines()
        version, status_code, reason = lines[0].split()
        headers = cls.parse_headers("\r\n".join(lines[1:]))
        return cls(
            version=version, status_code=status_code, reason=reason, headers=headers
        )

    def __str__(self):
        """Return complete SSDP response."""
        lines = list()
        lines.append(" ".join([self.version, str(self.status_code), self.reason]))
        for header in self.headers:
            lines.append("%s: %s" % header)
        return "\n".join(lines)

class SSDPRequest(SSDPMessage):
    """Simple Service Discovery Protocol (SSDP) request."""

    def __init__(self, method, uri="*", version="HTTP/1.1", headers=None):
        self.method = method
        self.uri = uri
        super().__init__(version=version, headers=headers)

    @classmethod
    def parse(cls, msg):
        """Parse message string to request object."""
        lines = msg.splitlines()
        method, uri, version = lines[0].split()
        headers = cls.parse_headers("\r\n".join(lines[1:]))
        return cls(version=version, uri=uri, method=method, headers=headers)

    def __str__(self):
        """Return complete SSDP request."""
        lines = list()
        lines.append(" ".join([self.method, self.uri, self.version]))
        for header in self.headers:
            lines.append("%s: %s" % header)
        return "\n".join(lines)

class SimpleServiceDiscoveryProtocol(asyncio.DatagramProtocol):
    """
    Simple Service Discovery Protocol (SSDP).
    SSDP is part of UPnP protocol stack. For more information see:
    https://en.wikipedia.org/wiki/Simple_Service_Discovery_Protocol
    """
    def __init__(self, recvq, excq):
        """
        @param recvq    - asyncio.Queue for new datagrams
        @param excq     - asyncio.Queue for exceptions
        """
        self._recvq = recvq
        self._excq = excq

        # Transports are connected at the time a connection is made.
        self._transport = None

    def connection_made(self, transport):
        self.transport = transport

    def datagram_received(self, data, addr):
        data = data.decode()

        if data.startswith("HTTP/"):
            response = SSDPResponse.parse(data)
            if (response.headers_dictionary['USN'].startswith("uuid:pnp-webserver-") or 
                response.headers_dictionary['USN'].startswith("uuid:pnp-scheduler-") or 
                response.headers_dictionary['USN'].startswith("uuid:pnp-scheduler201-") or 
                response.headers_dictionary['USN'].startswith("uuid:pnp-touchscreen-") or 
                response.headers_dictionary['USN'].startswith("uuid:pnp-myhomeserver1-") or 
                response.headers_dictionary['USN'].startswith("uuid:upnp-Basic gateway-") or 
                response.headers_dictionary['USN'].startswith("uuid:upnp-IPscenariomodule-") or 
                response.headers_dictionary['USN'].startswith("uuid:upnp-IPscenarioModule-")):

                self._recvq.put_nowait({"address": addr[0], "ssdp_location": response.headers_dictionary['LOCATION'], "ssdp_st": response.headers_dictionary['ST']})

    def error_received(self, exc):
        self._excq.put_nowait(exc)

    def connection_lost(self, exc):
        if exc is not None:
            self._excq.put_nowait(exc)

        if self._transport is not None:
            self._transport.close()
            self._transport = None

def _get_soap_body(ns: str, action: str) -> str:
    soap_body = f"""
        <?xml version="1.0"?>

        <soap:Envelope
        xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/"
        soap:encodingStyle="http://schemas.xmlsoap.org/soap/encoding">

        <soap:Body>
        <m:{action} xmlns:m="{ns}">
        </m:{action}>
        </soap:Body>

        </soap:Envelope>
    """
    return soap_body

async def get_port(SCPD_location: str) -> int:

    host = urlparse(SCPD_location).netloc
    scheme = urlparse(SCPD_location).scheme
    try:
        async with aiohttp.ClientSession() as session:
            
            service_ns = "urn:schemas-bticino-it:service:openserver:1"
            service_action = "getopenserverPort"
            service_control = "upnp/pwdControl"
            soap_body = _get_soap_body(service_ns, service_action)
            soap_action = f"{service_ns}#{service_action}"
            headers = {
                'SOAPAction': f'"{soap_action}"',
                'Host': f"{host}",
                'Content-Type': 'text/xml',
                'Content-Length': str(len(soap_body)),
            }

            ctrl_url = f"{scheme}://{host}/{service_control}"
            resp = await session.post(ctrl_url, data=soap_body, headers=headers)
            soap_response = xml.dom.minidom.parseString(await resp.text()).documentElement
            await session.close()

        return int(soap_response.getElementsByTagName("Port")[0].childNodes[0].data)
    except aiohttp.client_exceptions.ServerDisconnectedError:
        return 20000
    except aiohttp.client_exceptions.ClientOSError:
        return 20000

async def _get_scpd_details(SCPD_location: str) -> dict:

    discovery_info = dict()
    
    async with aiohttp.ClientSession() as session:
        scpd_response = await session.get(SCPD_location)
        scpd_xml = xml.dom.minidom.parseString(await scpd_response.text()).documentElement

        discovery_info["deviceType"] = scpd_xml.getElementsByTagName("deviceType")[0].childNodes[0].data
        discovery_info["friendlyName"] = scpd_xml.getElementsByTagName("friendlyName")[0].childNodes[0].data
        discovery_info["manufacturer"] = scpd_xml.getElementsByTagName("manufacturer")[0].childNodes[0].data
        discovery_info["manufacturerURL"] = scpd_xml.getElementsByTagName("manufacturerURL")[0].childNodes[0].data
        discovery_info["modelName"] = scpd_xml.getElementsByTagName("modelName")[0].childNodes[0].data
        discovery_info["modelNumber"] = scpd_xml.getElementsByTagName("modelNumber")[0].childNodes[0].data
        #discovery_info["presentationURL"] = scpd_xml.getElementsByTagName("presentationURL")[0].childNodes[0].data  ## bticino did not populate this field
        discovery_info["serialNumber"] = scpd_xml.getElementsByTagName("serialNumber")[0].childNodes[0].data
        discovery_info["UDN"] = scpd_xml.getElementsByTagName("UDN")[0].childNodes[0].data

        discovery_info["port"] = await get_port(SCPD_location)

        await session.close()

    return discovery_info

async def find_gateways() -> list:

    return_list = list()

    # Start the asyncio loop.
    loop = asyncio.get_running_loop()
    recvq = asyncio.Queue()
    excq = asyncio.Queue()

    search_request = bytes(SSDPRequest(
        "M-SEARCH",
        headers={
            "MX": "2",
            "ST": "upnp:rootdevice",
            "MAN": '"ssdp:discover"',
            "HOST": "239.255.255.250:1900",
            "Content-Length": "0",
        },
    ))

    transport, protocol = await loop.create_datagram_endpoint(lambda: SimpleServiceDiscoveryProtocol(recvq, excq), family=socket.AF_INET) # pylint: disable=unused-variable
    transport.sendto(search_request, ("239.255.255.250", 1900))
    try:
        await asyncio.sleep(2)
    finally:
        transport.close()

    while not recvq.empty():
        discovery_info = await recvq.get()
        discovery_info.update(await _get_scpd_details(discovery_info["ssdp_location"]))

        return_list.append(discovery_info)
    
    return return_list

async def get_gateway(address: str) -> dict:
    local_gateways = await find_gateways()
    for gateway in local_gateways:
        if gateway["address"] == address:
            return gateway

if __name__ == "__main__":
    local_gateways = asyncio.run(find_gateways())

    for gateway in local_gateways:
        print(f"Address: {gateway['address']}")
        print(f"Port: {gateway['port']}")
        print(f"Manufacturer: {gateway['manufacturer']}")
        print(f"Model: {gateway['modelName']}")
        print(f"Firmware: {gateway['modelNumber']}")
        print(f"Serial: {gateway['serialNumber']}")
        print()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="OWNd",
    version="0.7.34",
    author="anotherjulien",
    url="https://github.com/anotherjulien/OWNd",
    author_email="yetanotherjulien@gmail.com",
    description="Python interface for the OpenWebNet protocol",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=["aiohttp", "pytz", "python-dateutil"],
    python_requires='>=3.8',
)